from datetime import datetime, timedelta
import io
import json
import urllib.request

from bs4 import BeautifulSoup
from PIL import Image
import requests

from .crypto import encrypt_metadata, meta_audible_app, CertAuth
from .localization import localization


class Client:
    def __init__(self, username=None, password=None, filename=None, local="us"):
        self._local = localization.get(local)

        self._login_cookies = {}
        self._adp_token = str()
        self._access_token = str()
        self._refresh_token = str()
        self._device_private_key = str()
        self._expires = 0

        if (username and password):
            self.init_session(username, password)
            self.auth_register()
            if filename:
                self.to_json_file(filename)
                self.filename = filename

        elif filename:
            self.from_json_file(filename)
            self.filename = filename   

    @property
    def login_cookies(self):
        return self._login_cookies

    @property
    def adp_token(self):
        return self._adp_token

    @property
    def access_token(self):
        return self._access_token

    @property
    def refresh_token(self):
        return self._refresh_token

    @property
    def device_private_key(self):
        return self._device_private_key

    @property
    def expires(self):
        return datetime.fromtimestamp(self._expires) - datetime.utcnow()

    def expired(self):
        if datetime.fromtimestamp(self._expires) <= datetime.utcnow():
            return True
        else:
            return False
   
    def from_json_file(self, filename=None):
        filename = filename or self.filename
        with open(filename, "r") as infile:
            body = json.load(infile)

        self._login_cookies = body["login_cookies"]
        self._adp_token = body["adp_token"]
        self._access_token = body["access_token"]
        self._refresh_token = body["refresh_token"]
        self._device_private_key = body["device_private_key"]
        self._expires = body["expires"]

    def to_json_file(self, filename=None, indent=4):
        filename = filename or self.filename

        body = {}
        body["login_cookies"] = self._login_cookies
        body["adp_token"] = self._adp_token
        body["access_token"] = self._access_token
        body["refresh_token"] = self._refresh_token
        body["device_private_key"] = self._device_private_key
        body["expires"] = self._expires

        with open(filename, "w") as outfile:
            json.dump(body, outfile, indent=indent)

    def init_session(self, username, password):
        timeout = 30
        session = requests.Session()
        session.headers.update(
            {
                "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
                "Accept-Charset": "utf-8",
                "Accept-Language": self._local["Accept-Language"],
                "Host": self._local["AMAZON_LOGIN"].netloc,
                "Origin": self._local["AMAZON_LOGIN"].geturl(),
                "User-Agent": "Mozilla/5.0 (iPhone; CPU iPhone OS 12_3_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Mobile/15E148"
            }
        )

        while "session-token" not in session.cookies:
            session.get(self._local["AMAZON_LOGIN"].geturl(), timeout=timeout)

        oauth_params = {
            'openid.oa2.response_type': 'token',
            'openid.return_to': self._local["AMAZON_LOGIN"].geturl() + '/ap/maplanding',
            'openid.assoc_handle': self._local["openid_assoc_handle"],
            'openid.identity': 'http://specs.openid.net/auth/2.0/identifier_select',
            'pageId': 'amzn_audible_ios',
            'accountStatusPolicy': 'P1',
            'openid.claimed_id': 'http://specs.openid.net/auth/2.0/identifier_select',
            'openid.mode': 'checkid_setup',
            'openid.ns.oa2': 'http://www.amazon.com/ap/ext/oauth/2',
            'openid.oa2.client_id': 'device:6a52316c62706d53427a5735505a76477a45375959566674327959465a6374424a53497069546d45234132435a4a5a474c4b324a4a564d',
            'language': self._local["oauth_lang"],
            'openid.ns.pape': 'http://specs.openid.net/extensions/pape/1.0',
            'marketPlaceId': self._local["marketPlaceId"],
            'openid.oa2.scope': 'device_auth_access',
            'forceMobileLayout': 'true',
            'openid.ns': 'http://specs.openid.net/auth/2.0',
            'openid.pape.max_auth_age': '0'
        }
        response = session.get(self._local["AMAZON_LOGIN"].geturl() + "/ap/signin", params=oauth_params, timeout=timeout)
        oauth_url = response.request.url

        inputs = {}

        body = BeautifulSoup(response.text, "html.parser")
        for node in body.select("input[type=hidden]"):
            if node["name"] and node["value"]:
                inputs[node["name"]] = node["value"]
        inputs["email"] = username
        inputs["password"] = password

        metadata = meta_audible_app(session.headers.get("User-Agent"), oauth_url)
        inputs["metadata1"] = encrypt_metadata(metadata)

        session.headers.update(
            {
                "Referer": oauth_url,
                "Content-Type": "application/x-www-form-urlencoded"
            }
        )
        response = session.post(self._local["AMAZON_LOGIN"].geturl() + "/ap/signin", data=inputs, timeout=timeout)

        # check for captcha
        body = BeautifulSoup(response.text, "html.parser")
        captcha = body.find("img", alt=lambda x: x and "CAPTCHA" in x)

        while captcha:
            with urllib.request.urlopen(captcha["src"]) as url:
                f = io.BytesIO(url.read())

            img = Image.open(f)
            img.show()

            guess = input("Answer for CAPTCHA: ")
            guess = str(guess).strip().lower()

            inputs = {}
            for node in body.select("input[type=hidden]"):
                if node["name"] and node["value"]:
                    inputs[node["name"]] = node["value"]
            inputs["guess"] = guess
            inputs["use_image_captcha"] = "true"
            inputs["use_audio_captcha"] = "false"
            inputs["showPasswordChecked"] = "false"
            inputs["email"] = username
            inputs["password"] = password

            response = session.post(self._local["AMAZON_LOGIN"].geturl() + "/ap/signin", data=inputs, timeout=timeout)
        
            body = BeautifulSoup(response.text, "html.parser")
            captcha = body.find("img", alt=lambda x: x and "CAPTCHA" in x)

        # check for cvf identity check
        body = BeautifulSoup(response.text, "html.parser")
        cvf = body.find("div", id="cvf-page-content")

        if cvf:
            inputs = {}
            for node in body.select('input[type=hidden]'):
                if node["name"] and node["value"]:
                    inputs[node["name"]] = node["value"]
            headers = {
                "Content-Type": "application/x-www-form-urlencoded",
                "Referer": response.url
            }
            response = session.post(self._local["AMAZON_LOGIN"].geturl() + "/ap/cvf/verify", headers=headers, data=inputs, timeout=timeout)
    
            body = BeautifulSoup(response.text, "html.parser")
            inputs = {}
            for node in body.select('input[type=hidden]'):
                if node["name"] and node["value"]:
                    inputs[node["name"]] = node["value"]
            headers = {
                "Content-Type": "application/x-www-form-urlencoded",
                "Referer": response.url
            }
            inputs["action"] = "code"
            inputs["code"] = input("Code: ")
            response = session.post(self._local["AMAZON_LOGIN"].geturl() + "/ap/cvf/verify", headers=headers, data=inputs, timeout=timeout)

        if response.status_code == 404:
            map_landing = urllib.parse.parse_qs(response.url)
            self._access_token = map_landing["openid.oa2.access_token"][0]
            for cookie in session.cookies:
                self._login_cookies[cookie.name] = cookie.value
        else:
            raise Exception("Unable to login")

    def auth_register(self):
        json_cookies = []
        request_cookies = {}
        for key, val in self._login_cookies.items():
            json_cookies.append({
                "Name": key,
                "Value": val.replace(r'"', r'')
            })
            request_cookies[key] = val.replace(r'"', r'')

        json_object = {
            "requested_token_type": ["bearer", "mac_dms", "website_cookies"],
            "cookies": {
                "website_cookies": json_cookies,
                "domain": self._local["auth_register_domain"]
            },
            "registration_data": {
                "domain": "Device",
                "app_version": "3.7",
                "device_serial": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
                "device_type": "A2CZJZGLK2JJVM",
                "device_name": "%FIRST_NAME%%FIRST_NAME_POSSESSIVE_STRING%%DUPE_STRATEGY_1ST%Audible for iPhone",
                "os_version": "12.3.1",
                "device_model": "iPhone",
                "app_name": "Audible"
            },
            "auth_data": {
                "access_token": self._access_token
            },
            "requested_extensions": ["device_info", "customer_info"]
        }

        headers = {
            "Host": self._local["AMAZON_API"].netloc,
            "Content-Type": "application/json",
            "Accept-Charset": "utf-8",
            "x-amzn-identity-auth-domain": self._local["AMAZON_API"].netloc,
            "Accept": "application/json",
            "User-Agent": "AmazonWebView/Audible/3.7/iOS/12.3.1/iPhone",
            "Accept-Language": self._local["Accept-Language"]
        }

        url = self._local["AMAZON_API"].geturl() + "/auth/register"
        response = requests.post(url, json=json_object, headers=headers, cookies=request_cookies)

        body = response.json()
        if response.status_code != 200:
            raise Exception(body)

        tokens = body["response"]["success"]["tokens"]

        self._adp_token = tokens["mac_dms"]["adp_token"]
        self._device_private_key = tokens["mac_dms"]["device_private_key"]

        self._access_token = tokens["bearer"]["access_token"]
        self._refresh_token = tokens["bearer"]["refresh_token"]
        self._expires = (datetime.utcnow() + timedelta(seconds=int(tokens["bearer"]["expires_in"]))).timestamp()

        website_cookies = tokens["website_cookies"]
        for cookie in website_cookies:
            self._login_cookies[cookie["Name"]] = cookie["Value"]

    def auth_deregister(self):
        json_object = {
            "deregister_all_existing_accounts": "true"
        }
        headers = {
            "Host": self._local["AMAZON_API"].netloc,
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept-Charset": "utf-8",
            "x-amzn-identity-auth-domain": self._local["AMAZON_API"].netloc,
            "Accept": "application/json",
            "User-Agent": "AmazonWebView/Audible/3.7/iOS/12.3.1/iPhone",
            "Accept-Language": self._local["Accept-Language"],
            "Authorization": f"Bearer {self._access_token}"
        }        

        request_cookies = {}
        for key, val in self._login_cookies.items():
            request_cookies[key] = val.replace(r'"', r'')

        url = self._local["AMAZON_API"].geturl() + "/auth/deregister"
        response = requests.post(url, json=json_object, headers=headers, cookies=request_cookies)

        body = response.json()
        if response.status_code != 200:
            raise Exception(body)
        self._adp_token = str()
        self._refresh_token = str()
        self._expires = 0

    def refresh_access_token(self, force=False):
        if force or self.expired():
            body = {
                "app_name": "Audible",
                "app_version": "3.7",
                "source_token": self._refresh_token,
                "requested_token_type": "access_token",
                "source_token_type": "refresh_token"
            }
    
            headers = {
                "Content-Type": "application/x-www-form-urlencoded",
                "x-amzn-identity-auth-domain": self._local["AMAZON_API"].netloc
            }
    
            url = self._local["AMAZON_API"].geturl() + "/auth/token"
            response = requests.post(url, data=body, headers=headers)
    
            body = response.json()
            if response.status_code != 200:
                raise Exception(body)
            self._access_token = body["access_token"]
            self._expires = (datetime.utcnow() + timedelta(seconds=int(body["expires_in"]))).timestamp()

    def user_profile(self):
        headers = {
            "Host": self._local["AMAZON_API"].netloc,
            "Accept-Charset": "utf-8",
            "User-Agent": "AmazonWebView/Audible/3.7/iOS/12.3.1/iPhone",
            "Accept-Language": self._local["Accept-Language"],
            "Authorization": f"Bearer {self._access_token}"
        }

        request_cookies = {}
        for key, val in self._login_cookies.items():
            request_cookies[key] = val.replace(r'"', r'')

        url = self._local["AMAZON_API"].geturl() + "/user/profile"
        response = requests.get(url, headers=headers, cookies=request_cookies)

        return response.json()

    def refresh_or_register(self, force=False):
        try:
            self.refresh_access_token(force=force)
        except:
            try:
                self.auth_deregister()
                self.auth_register()
            except:
                raise Exception("Could not refresh client.")

    def _api_request(self, method, path, api_version="1.0", json_data=None, auth=None, cookies=None, **params):
        url = "/".join((self._local["AUDIBLE_API"].geturl(), api_version, path))
        headers = {
            "Accept": 'application/json',
            'Content-Type': 'application/json'
        }
        auth = CertAuth(self._adp_token, self._device_private_key)
        resp = requests.request(
            method, url, json=json_data, params=params, headers=headers, auth=auth, cookies=cookies
        )
        resp.raise_for_status()

        return resp.json()

    def get(self, path, **params):
        return self._api_request("GET", path, **params)

    def post(self, path, body, **params):
        return self._api_request("POST", path, json_data=body, **params)

    def delete(self, path, **params):
        return self._api_request("DELETE", path, **params)




from urllib.parse import urlparse


localization = {
    "de": {
        "AMAZON_LOGIN": urlparse("https://www.amazon.de"),
        "AMAZON_API": urlparse("https://api.amazon.de"),
        "AUDIBLE_API": urlparse("https://api.audible.de"),
        "Accept-Language": "de-DE",
        "marketPlaceId": "AN7V1F1VY261K",
        "openid_assoc_handle": "amzn_audible_ios_de",
        "oauth_lang": "de-DE",
        "auth_register_domain": ".amazon.de"
    },
    "us": {
        "AMAZON_LOGIN": urlparse("https://www.amazon.com"),
        "AMAZON_API": urlparse("https://api.amazon.com"),
        "AUDIBLE_API": urlparse("https://api.audible.com"),
        "Accept-Language": "en-US",
        "marketPlaceId": "AF2M0KC94RCEA",
        "openid_assoc_handle": "amzn_audible_ios_us",
        "oauth_lang": "en-US",
        "auth_register_domain": ".amazon.com"
    },
    "uk": {
        "AMAZON_LOGIN": urlparse("https://www.amazon.co.uk"),
        "AMAZON_API": urlparse("https://api.amazon.co.uk"),
        "AUDIBLE_API": urlparse("https://api.audible.co.uk"),
        "Accept-Language": "en-GB",
        "marketPlaceId": "A2I9A3Q2GNFNGQ",
        "openid_assoc_handle": "amzn_audible_ios_uk",
        "oauth_lang": "en-GB",
        "auth_register_domain": ".amazon.co.uk"
    },
    "fr": {
        "AMAZON_LOGIN": urlparse("https://www.amazon.fr"),
        "AMAZON_API": urlparse("https://api.amazon.fr"),
        "AUDIBLE_API": urlparse("https://api.audible.fr"),
        "Accept-Language": "fr-FR",
        "marketPlaceId": "A2728XDNODOQ8T",
        "openid_assoc_handle": "amzn_audible_ios_fr",
        "oauth_lang": "fr-FR",
        "auth_register_domain": ".amazon.fr"
    }
}




import base64
import binascii
import ctypes
from datetime import datetime
import math
from urllib.parse import urlparse

from requests.auth import AuthBase
import rsa


def encrypt_metadata(metadata: str) -> str:
    """
    This function encrypt given metadata. Encrypted metadata are required for login to amazon.

    :param metadata: plaintext metadata
    :type metadata: str
    :return: encrypted metadata
    :rtype: str
    """
    checksum = binascii.crc32(metadata.encode()) & 0xffffffff
    checksum = checksum.to_bytes(4, byteorder='big').hex().upper()

    object_bytes = f"{checksum}#{metadata}".encode()

    temp2 = []
    for i in range(0, len(object_bytes), 4):
        data = int.from_bytes(object_bytes[i:i+4], 'little')
        temp2.append(data)

    wrap_constant = 2654435769
    constants = [1888420705, 2576816180, 2347232058, 874813317]

    rounds = math.ceil(len(object_bytes) / 4)
    minor_rounds = math.floor(6 + (52 / rounds))
    first = temp2[0]
    last = temp2[rounds - 1]

    inner_roll = 0
    while minor_rounds > 0:
        minor_rounds -= 1

        inner_roll += wrap_constant
        inner_variable = inner_roll >> 2 & 3

        for i in range(rounds):
            first = temp2[(i + 1) % rounds]
            temp2[i] += ((last >> 5 ^ first << 2)
                         + (first >> 3 ^ last << 4) ^ (inner_roll ^ first)
                         + (constants[i & 3 ^ inner_variable] ^ last))
            temp2[i] = ctypes.c_uint32(temp2[i]).value
            last = temp2[i]

    final_round = []
    for i in range(rounds):
        data2 = temp2[i].to_bytes(4, 'little')
        final_round.append(data2)

    final_round = b"".join(final_round)
    base64_encoded = base64.standard_b64encode(final_round)

    final = f"ECdITeCs:{base64_encoded.decode()}"

    return final


def decrypt_metadata(metadata: str) -> str:
    """
    This function decrypt given metadata. Decrypt metadata is for testing purposes only.

    :param metadata: encrypted metadata
    :type metadata: str
    :return: plaintext metadata
    :rtype: str
    """
    metadata = metadata.lstrip("ECdITeCs:")

    final_round = base64.b64decode(metadata)
    temp2 = []

    for i in range(0, len(final_round), 4):
        data = int.from_bytes(final_round[i:i+4], 'little')
        temp2.append(data)

    rounds = len(temp2)
    minor_rounds = math.floor(6 + (52 / rounds))

    wrap_constant = 2654435769
    constants = [1888420705, 2576816180, 2347232058, 874813317]

    inner_roll = 0
    inner_variable = 0

    for _ in range(minor_rounds + 1):
        inner_roll += wrap_constant
        inner_variable = inner_roll >> 2 & 3

    while minor_rounds > 0:
        minor_rounds -= 1

        inner_roll -= wrap_constant
        inner_variable = inner_roll >> 2 & 3

        for i in range(rounds):
            i = rounds - i - 1

            first = temp2[(i + 1) % rounds]
            last = temp2[(i - 1) % rounds]

            temp2[i] -= ((last >> 5 ^ first << 2)
                         + (first >> 3 ^ last << 4) ^ (inner_roll ^ first)
                         + (constants[i & 3 ^ inner_variable] ^ last))
            temp2[i] = ctypes.c_uint32(temp2[i]).value
            last = temp2[i]

    object_bytes = []
    for block in temp2:
        for align in {0, 8, 16, 24}:
            if block >> align & 255 != 0:
                object_bytes.append(chr(block >> align & 255))

    object_bytes = "".join(object_bytes)

    return object_bytes[9:]


class CertAuth(AuthBase):
    def __init__(self, adp_token, device_cert):
        self.adp_token = adp_token
        self.device_cert = device_cert

    def __call__(self, r):
        url_parsed = urlparse(r.url)
        path = url_parsed[2]
        query = url_parsed[4]

        url = path[:]
        if query:
            url += f"?{query}"

        headers = sign_request(url, r.method, r.body, self.adp_token, self.device_cert)
        for item in headers:
            r.headers[item] = headers[item]

        return r


def sign_request(url: str, method: str, body: str, adp_token: str,
                 private_key: str) -> dict:
    """
    Helper function who create a signed header params.

    :param url: the requested url
    :param method: the request method (GET, POST, DELETE, ...)
    :param body: the http message body
    :param adp_token: the token is obtained after register as device
    :param private_key: the rsa key obtained after register as device
    :type url: str
    :type method: str
    :type body: str
    :type adp_token: str
    :type private_key: str
    :return:
    :rtype: dict
    """
    date = datetime.utcnow().isoformat("T") + "Z"

    data = f"{method}\n{url}\n{date}\n"
    if body:
        data += body
    data += "\n"
    data += adp_token

    key = rsa.PrivateKey.load_pkcs1(private_key)
    cipher = rsa.pkcs1.sign(data.encode(), key, "SHA-256")
    signed_encoded = base64.b64encode(cipher)

    signature = f"{signed_encoded.decode()}:{date}"

    return {
        "x-adp-token": adp_token,
        "x-adp-alg": "SHA256withRSA:1.0",
        "x-adp-signature": signature
    }


def now_to_unix_ms() -> int:
    return math.floor(datetime.now().timestamp()*1000)


def meta_audible_app(user_agent: str, oauth_url: str) -> str:
    """
    :param user_agent: user agent used for request
    :param oauth_url: referer location url
    :type user_agent: str
    :type oauth_url: str
    :return: plaintext metadata simulating sign-in from audible ios app
    :rtype: str
    """
    return f'{{"start":{now_to_unix_ms()},"interaction":{{"keys":0,"keyPressTimeIntervals":[],"copies":0,"cuts":0,"pastes":0,"clicks":0,"touches":0,"mouseClickPositions":[],"keyCycles":[],"mouseCycles":[],"touchCycles":[]}},"version":"3.0.0","lsUbid":"X39-6721012-8795219:1549849158","timeZone":-6,"scripts":{{"dynamicUrls":["https://images-na.ssl-images-amazon.com/images/I/61HHaoAEflL._RC|11-BZEJ8lnL.js,01qkmZhGmAL.js,71qOHv6nKaL.js_.js?AUIClients/AudibleiOSMobileWhiteAuthSkin#mobile","https://images-na.ssl-images-amazon.com/images/I/21T7I7qVEeL._RC|21T1XtqIBZL.js,21WEJWRAQlL.js,31DwnWh8lFL.js,21VKEfzET-L.js,01fHQhWQYWL.js,51TfwrUQAQL.js_.js?AUIClients/AuthenticationPortalAssets#mobile","https://images-na.ssl-images-amazon.com/images/I/0173Lf6yxEL.js?AUIClients/AuthenticationPortalInlineAssets","https://images-na.ssl-images-amazon.com/images/I/211S6hvLW6L.js?AUIClients/CVFAssets","https://images-na.ssl-images-amazon.com/images/G/01/x-locale/common/login/fwcim._CB454428048_.js"],"inlineHashes":[-1746719145,1334687281,-314038750,1184642547,-137736901,318224283,585973559,1103694443,11288800,-1611905557,1800521327,-1171760960,-898892073],"elapsed":52,"dynamicUrlCount":5,"inlineHashesCount":13}},"plugins":"unknown||320-568-548-32-*-*-*","dupedPlugins":"unknown||320-568-548-32-*-*-*","screenInfo":"320-568-548-32-*-*-*","capabilities":{{"js":{{"audio":true,"geolocation":true,"localStorage":"supported","touch":true,"video":true,"webWorker":true}},"css":{{"textShadow":true,"textStroke":true,"boxShadow":true,"borderRadius":true,"borderImage":true,"opacity":true,"transform":true,"transition":true}},"elapsed":1}},"referrer":"","userAgent":"{user_agent}","location":"{oauth_url}","webDriver":null,"history":{{"length":1}},"gpu":{{"vendor":"Apple Inc.","model":"Apple A9 GPU","extensions":[]}},"math":{{"tan":"-1.4214488238747243","sin":"0.8178819121159085","cos":"-0.5753861119575491"}},"performance":{{"timing":{{"navigationStart":{now_to_unix_ms()},"unloadEventStart":0,"unloadEventEnd":0,"redirectStart":0,"redirectEnd":0,"fetchStart":{now_to_unix_ms()},"domainLookupStart":{now_to_unix_ms()},"domainLookupEnd":{now_to_unix_ms()},"connectStart":{now_to_unix_ms()},"connectEnd":{now_to_unix_ms()},"secureConnectionStart":{now_to_unix_ms()},"requestStart":{now_to_unix_ms()},"responseStart":{now_to_unix_ms()},"responseEnd":{now_to_unix_ms()},"domLoading":{now_to_unix_ms()},"domInteractive":{now_to_unix_ms()},"domContentLoadedEventStart":{now_to_unix_ms()},"domContentLoadedEventEnd":{now_to_unix_ms()},"domComplete":{now_to_unix_ms()},"loadEventStart":{now_to_unix_ms()},"loadEventEnd":{now_to_unix_ms()}}}}},"end":{now_to_unix_ms()},"timeToSubmit":108873,"form":{{"email":{{"keys":0,"keyPressTimeIntervals":[],"copies":0,"cuts":0,"pastes":0,"clicks":0,"touches":0,"mouseClickPositions":[],"keyCycles":[],"mouseCycles":[],"touchCycles":[],"width":290,"height":43,"checksum":"C860E86B","time":12773,"autocomplete":false,"prefilled":false}},"password":{{"keys":0,"keyPressTimeIntervals":[],"copies":0,"cuts":0,"pastes":0,"clicks":0,"touches":0,"mouseClickPositions":[],"keyCycles":[],"mouseCycles":[],"touchCycles":[],"width":290,"height":43,"time":10353,"autocomplete":false,"prefilled":false}}}},"canvas":{{"hash":-373378155,"emailHash":-1447130560,"histogramBins":[]}},"token":null,"errors":[],"metrics":[{{"n":"fwcim-mercury-collector","t":0}},{{"n":"fwcim-instant-collector","t":0}},{{"n":"fwcim-element-telemetry-collector","t":2}},{{"n":"fwcim-script-version-collector","t":0}},{{"n":"fwcim-local-storage-identifier-collector","t":0}},{{"n":"fwcim-timezone-collector","t":0}},{{"n":"fwcim-script-collector","t":1}},{{"n":"fwcim-plugin-collector","t":0}},{{"n":"fwcim-capability-collector","t":1}},{{"n":"fwcim-browser-collector","t":0}},{{"n":"fwcim-history-collector","t":0}},{{"n":"fwcim-gpu-collector","t":1}},{{"n":"fwcim-battery-collector","t":0}},{{"n":"fwcim-dnt-collector","t":0}},{{"n":"fwcim-math-fingerprint-collector","t":0}},{{"n":"fwcim-performance-collector","t":0}},{{"n":"fwcim-timer-collector","t":0}},{{"n":"fwcim-time-to-submit-collector","t":0}},{{"n":"fwcim-form-input-telemetry-collector","t":4}},{{"n":"fwcim-canvas-collector","t":2}},{{"n":"fwcim-captcha-telemetry-collector","t":0}},{{"n":"fwcim-proof-of-work-collector","t":1}},{{"n":"fwcim-ubf-collector","t":0}},{{"n":"fwcim-timer-collector","t":0}}]}}'

